#include <fits.h>
#include "wfpc2psfdata.h"
#include "wfpc2distort.h"

ftype fits,dq;
double GAIN,RN,EXP,EXP0,EPOCH,FIXED_ET=-1.;
float DMIN,DMAX;
int MASKCR=1;

int WFPC2type(ftype *f) {
   int i;
   if (f->img.X==800 && f->img.Y==800 && f->img.Z==4) return 1;
   if (f->Next==4) {
      for (i=0;i<4 && f->ext[i].X==800 && f->ext[i].Y==800 && f->ext[i].Z==1;i++);
      if (i==4) return 2;
   }
   if (f->Next==3 && f->ext[0].X>1200 && f->ext[0].Y>1200 && f->ext[0].Z==1) {
      for (i=1;i<3 && f->ext[i].X==f->ext[0].X && f->ext[i].Y==f->ext[0].Y && f->ext[i].Z==1;i++);
      if (i==3) return -1;
   }
   return 0;
}

void WFPC2getcards(int drz) {
   if (strcmp(getcardval(&(fits.img),"FILETYPE",1),"SCI") || strcmp(getcardval(&(fits.img),"TELESCOP",1),"HST") || strcmp(getcardval(&(fits.img),"INSTRUME",1),"WFPC2")) {printf("**Format error (filetype,telescop,instrume)\n"); exit(-1);}
   GAIN=atof(getcardval(&(fits.img),"ATODGAIN",1));
   if (GAIN==7) RN=5.0;
   else if (GAIN==14) RN=7.5;
   else if (GAIN==15) {
      RN=7.5;
      GAIN=14;
   }
   else {printf("**Format error (atodgain)\n"); exit(-1);}
   if (FIXED_ET>0) EXP=FIXED_ET;
   else EXP=atof(getcardval(&(fits.img),"EXPTIME",1));
   DMIN=(float)atof(getcardval(&(fits.img),"RSDPFILL",1))-1.;
   DMAX=(float)atof(getcardval(&(fits.img),"SATURATE",1))+1.;
   EPOCH=0.5*(atof(getcardval(&(fits.img),"EXPSTART",1))+atof(getcardval(&(fits.img),"EXPEND",1)));
   EXP0=EXP;
   return;
}

void WFPC2toextensions(void) {
   int ext;
   if (fits.Next>0) {
      for (ext=0;ext<fits.Next;ext++) freeim(fits.ext+ext);
      free(fits.ext);
   }
   fits.Next=4;
   fits.ext=(imtype*)calloc(sizeof(imtype),fits.Next);
   if (!fits.ext) merr();
   for (ext=0;ext<4;ext++) {
      fits.ext[ext].Ncards=0;
      fits.ext[ext].Nmax=0;
      fits.ext[ext].X=fits.ext[ext].Y=800;
      fits.ext[ext].Z=1;
      fits.ext[ext].bits=-32;
      fits.ext[ext].pcount=0;
      fits.ext[ext].bzero=0.;
      fits.ext[ext].bscale=1.;
      strcpy(fits.ext[ext].xtension,"IMAGE");
      fits.ext[ext].cards=NULL;
      fits.ext[ext].data=allocimg(800,800,1);
      memcpy(fits.ext[ext].data[0][0],fits.img.data[ext][0],sizeof(float)*800*800);
   }
   freeimg(fits.img.data,fits.img.X,fits.img.Y,fits.img.Z);
   fits.img.X=fits.img.Y=fits.img.Z=0;
   fits.img.bits=16;
   fits.img.data=allocimg(0,0,0);
   return;
}

void WFPC2mask(int drz,int tp) {
   int x,y,z,idq;
   for (z=0;z<fits.Next;z++) for (y=0;y<fits.ext[z].Y;y++) for (x=0;x<fits.ext[z].X;x++) {
      if (tp==1) idq = dq.img.data[z][y][x];
      else idq = dq.ext[z].data[0][y][x];
      if (idq&8 || fits.ext[z].data[0][y][x]>3500) fits.ext[z].data[0][y][x]=DMAX+1;
      else if (idq&951 || y==fits.ext[z].Y-1 || x==fits.ext[z].X-1) fits.ext[z].data[0][y][x]=DMIN-1; // flags 1, 2, 4, 16, 32, 128, 256, 512
      if (idq&6208) printf("Illegal data quality value %d %d %d\n",idq&64,idq&2048,idq&4096); // flags 64, 2048, 4096 (no 1024 check??)
   }
   return;
}

void WFPC2pixcorr(void) {
   int x,y,z;
   float DMIN1,DMAX1;

   if (DMAX>0.) DMAX1=1.5*DMAX;
   else DMAX1=0.;
   if (DMIN<0.) DMIN1=1.5*DMIN;
   else DMIN1=0.;
   for (z=0;z<fits.Next;z++) for (y=0;y<fits.ext[z].Y;y++) for (x=0;x<fits.ext[z].X;x++) if (fits.ext[z].data[0][y][x]>DMIN && fits.ext[z].data[0][y][x]<DMAX) {
      fits.ext[z].data[0][y][x]*=WFPC2xsize(z,x,y)*WFPC2ysize(z,x,y);
   }
   else if (fits.ext[z].data[0][y][x]>=DMAX) fits.ext[z].data[0][y][x]=DMAX1+1;
   else fits.ext[z].data[0][y][x]=DMIN1-1;
   DMIN=DMIN1;
   DMAX=DMAX1;
   return;
}

void WFPC2drzpixcorr(double mult) {
   int x,y;

   for (y=0;y<fits.ext[0].Y;y++) for (x=0;x<fits.ext[0].X;x++) {
      if (fits.ext[1].data[0][y][x]==0.) fits.ext[0].data[0][y][x]=DMIN-1; // no weight
   }
   mult*=GAIN;
   DMAX*=EXP/mult;
   DMIN*=EXP/mult;
   for (y=0;y<fits.ext[0].Y;y++) for (x=0;x<fits.ext[0].X;x++) {
      fits.ext[0].data[0][y][x]*=EXP/mult;
   }
   freeim(fits.ext+1);
   freeim(fits.ext+2);
   fits.Next-=2;
   return;
}

void WFPC2setcards(int tp) {
   int ext;
   if (tp==-1) {
      insertcards(fits.ext,GAIN,RN,EXP,DMIN,DMAX,EPOCH,0.0,EXP0);
      return;
   }
   for (ext=0;ext<4;ext++) {
      insertcards(fits.ext+ext,GAIN,RN,EXP,DMIN,DMAX,EPOCH,0.0,EXP0);
   }
   return;
}

int main(int argc,char**argv) {
   int i,tp;
   //int fcorr=0;

   if (argc<2) {
      printf("Usage: %s <<-flags>> <fits file, data quality file>\n",*argv);
      printf(" -keepcr      leaves fixed cosmic ray pixels in image\n");
      //printf(" -flashdir=X  sets directory prefix for flash correction files\n");
      return 1;
   }
   for (i=1;i<argc;i++) if (!strcasecmp(argv[i],"-KEEPCR")) MASKCR=0;
   else if (!strncasecmp(argv[i],"-EXPTIME=",9)) FIXED_ET=atof(argv[i]+9);
   //else if (!strncasecmp(argv[i],"-FLASHDIR=",10)) strcpy(FLASHDIR,argv[i]+10);
   else {
      readfits(argv[i],&fits,1);
      tp=WFPC2type(&fits);
      if (strcmp(getcardval(&fits.img,"DOLWFPC2",0),"")) printf("%s already run through wfpc2mask\n",argv[i]);
      else if (tp==0) printf("%s is not a WFPC2 fits file\n",argv[i]);
      else {
	 if (tp>0) {
	    if (i==argc-1) {
	       printf("No data quality file listed for %s\n",argv[i]);
	       return 1;
	    }
	    readfits(argv[i+1],&dq,1);
	    if (tp!=WFPC2type(&dq)) printf("%s and %s have different image types\n",argv[i],argv[i+1]);
	 }
	 WFPC2getcards(0);
	 if (tp==1) {
	    WFPC2toextensions();
	 }
	 if (tp>0) {
	    WFPC2mask(0,tp);
	    WFPC2pixcorr();
	 }
	 else WFPC2drzpixcorr(1.0);
	 WFPC2setcards(tp);
	 if (tp>0) {
	    addcard(fits.ext,"DOLWFPC2=                    0 / DOLPHOT WFPC2 tag                              ");
	    addcard(fits.ext+1,"DOLWFPC2=                    1 / DOLPHOT WFPC2 tag                              ");
	    addcard(fits.ext+2,"DOLWFPC2=                    2 / DOLPHOT WFPC2 tag                              ");
	    addcard(fits.ext+3,"DOLWFPC2=                    3 / DOLPHOT WFPC2 tag                              ");
	 }
	 else {
	    addcard(fits.ext,"DOLWFPC2=                   -1 / DOLPHOT WFPC2 tag                              ");
	 }
	 writefits(argv[i],&fits,1);
      }
      //if (fcorr) freefits(&flash);
      freefits(&fits);
      if (tp>0) {
	 freefits(&dq);
	 i++;  // skip DQ image
      }
   }
   return 0;
}
