#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fits.h>
#include "wfpc2psfdata.h"
#include "wfpc2filters.h"
#include "wfpc2distort.h"

typedef float apskytype[2];
typedef double apsftype[3][6];
typedef float dpostype[5];
extern apsftype *apsf;
extern apskytype *apsky;
extern imtype *datahd,*dataim;
extern int Timg,Nimg;
extern int *rphot,*RPSF,SubPixel,FPSF,EPSF;
extern double *RAper,MinS,MaxS,MaxE,PSFStep,Zero,*apsize,*RSky0,*RSky1;
extern int PSFsol;

typedef int int4[4];
static int4 *wfpc2psflu=NULL;
static float********wfpc2psflib=NULL;
int WFPC2_DRIZZLE_BASE=0;
extern int FlagMask,UseCTE;
static int WFPC2_RPSF=34;

extern int DEBUG;
extern double *iGAIN,*iEXP,*iEXP0,*iEPOCH,*apcor;
extern void shift(int img,float x0,float y0,float *x,float *y,int dir);
extern dpostype *dpos,*dpos0;

void wfpc2initparam(void) {
   int img;
   FPSF=4;
   SubPixel=1;
   EPSF=0;
   PSFsol=-1;
   PSFStep=0.;
   Zero=0.;
   MinS=0.2;
   MaxS=15.;
   MaxE=0.1;
   for (img=0;img<Timg;img++) {
      if (RPSF[img]>=wfpc2_rpsf[0]) {RPSF[img]=wfpc2_rpsf[0]-1; printf("Lowering RPSF to %d\n",RPSF[img]);}
      if (RAper[img]>wfpc2_rpsf[0]-1.) {RAper[img]=wfpc2_rpsf[0]-1; printf("Lowering RAper to %d\n",wfpc2_rpsf[0]-1);}
   }
   return;
}

void wfpc2radii(int setShift) {
   char str[81],*ptr;
   int img;
   if (wfpc2psflu==NULL) {
      wfpc2psflu=(int4*)calloc(sizeof(int4),Timg);
      if (!wfpc2psflu) merr();
   }
   for (img=0;img<Timg;img++) {
      strcpy(str,getcardval(dataim+img,"DOLWFPC2",0));
      wfpc2psflu[img][0]=strtol(str,&ptr,10);
      if (ptr==str) {
	 printf("**Image %d has not been preprocessed with wfpc2mask; cannot proceed\n",img+1);
	 exit(-1);
      }
      else if (wfpc2psflu[img][0]==-1 && img<Nimg) {
	 printf("**Image %d has been drizzled; cannot run photometry\n",img+1);
	 exit(-1);
      }
      else if (wfpc2psflu[img][0]<-1 || wfpc2psflu[img][0]>3) {
	 printf("**Image %d's chip cannot be identified; please report bug.\n",img+1);
	 exit(-1);
      }
      // use PC PSFs for drizzled frame
      if (wfpc2psflu[img][0]==-1) {
	 wfpc2psflu[img][0]=0;
	 WFPC2_DRIZZLE_BASE=1;
      }
      if (wfpc2psflu[img][0]>0) {
	 RAper[img]/=1.5;
	 RSky0[img]/=1.5;
	 RSky1[img]/=1.5;
	 RPSF[img]=(RPSF[img]*2+2)/3;
	 apsize[img]/=1.5;
	 apsky[img][0]/=1.5;
	 apsky[img][1]/=1.5;
	 if (setShift) {
	    switch(wfpc2psflu[img][0]) {
	    case 1:
	       dpos[img][0]= 0.457*dpos0[img][1];
	       dpos[img][1]=-0.457*dpos0[img][0];
	       break;
	    case 2:
	       dpos[img][0]=-0.457*dpos0[img][0];
	       dpos[img][1]=-0.457*dpos0[img][1];
	       break;
	    case 3:
	       dpos[img][0]=-0.457*dpos0[img][1];
	       dpos[img][1]= 0.457*dpos0[img][0];
	       break;
	    }
	 }
      }
   }
}

void wfpc2initpsf(void) {
   int img,i,j,y1,x1,y2,x2,y3,n2,n3,n3skip;
   char str[81];
   FILE *f;
   float***** ptr1;
   float**** ptr2;
   float*** ptr3;
   float** ptr4;
   float* ptr5;
   float Temp;

   if (WFPC2_NFILTERS<0) WFPC2initfilters();
   if (wfpc2psflib==NULL) {
      wfpc2psflib=(float********)calloc(sizeof(float*******),3);
      if (!wfpc2psflib) merr();
      for (i=0;i<4;i++) {
	 wfpc2psflib[i]=(float*******)calloc(sizeof(float******),WFPC2_NFILTERS);
	 if (!wfpc2psflib[i]) merr();
      }
   }
   for (i=0;i<3;i++) for (j=0;j<WFPC2_NFILTERS;j++) wfpc2psflib[i][j]=NULL;
   for (img=0;img<Timg;img++) {
      strcpy(str,getcardval(datahd+img,"FILTNAM1",0));
      if (!strcmp(str,"")) strcpy(str,getcardval(datahd+img,"FILTNAM2",0));
      wfpc2psflu[img][1]=WFPC2findfilt(str);
      //printf("Image %d: cm=%d, filt=%d (%s)\n",img+1,wfpc2psflu[img][0],wfpc2psflu[img][1],WFPC2filters[wfpc2psflu[img][1]].name);
      if (RPSF[img]>=wfpc2_rpsf[wfpc2psflu[img][0]]) {printf("ERROR: RPSF must be less than %d for WFPC2/WFC data\n",wfpc2_rpsf[wfpc2psflu[img][0]]); exit(-1);}
      if (RAper[img]>wfpc2_rpsf[wfpc2psflu[img][0]]-1.) {printf("ERROR: RAper must be no more than %d for WFPC2/WFC data\n",wfpc2_rpsf[wfpc2psflu[img][0]]-1); exit(-1);}
      // Extract temperature
      Temp=atof(getcardval(datahd+img,"UCH1CJTM",1));
      if (Temp<-82) wfpc2psflu[img][2]=0;
      else wfpc2psflu[img][2]=1;
      // Extract gain
      if (fabs(iGAIN[img]-7.)<0.001) {
	 wfpc2psflu[img][3]=0;
	 iGAIN[img]=7.;
      }
      else if (fabs(iGAIN[img]-14.)<0.001 || fabs(iGAIN[img]-15.)<0.001) {
	 wfpc2psflu[img][3]=1;
	 iGAIN[img]=14.;
      }
      else {
	 printf("ERROR: WFPC2 GAIN must be 7 or 14.\n");
      }
   }
   WFPC2_RPSF=0;
   for (img=0;img<Timg;img++) {
      i=wfpc2psflu[img][0];
      apsf[img][0][0]=1.;
      apsf[img][1][0]=1.;
      apsf[img][2][0]=0.;
      if (i==0) apsize[img]=11.;
      else apsize[img]=5.;
      for (i=1;i<5;i++) apsf[img][0][i]=apsf[img][1][i]=apsf[img][2][i]=0.;
      if (WFPC2_RPSF<RPSF[img]) WFPC2_RPSF=RPSF[img];
      if (WFPC2_RPSF<rphot[img]) WFPC2_RPSF=rphot[img];
   }
   for (img=0;img<Timg;img++) if (wfpc2psflib[i=wfpc2psflu[img][0]][j=wfpc2psflu[img][1]]==NULL) {
      sprintf(str,"%s/wfpc2/data/%s.%s.psf",BASEDIR,WFPC2filters[wfpc2psflu[img][1]].name,wfpc2_cn[i]);
      if ((f=fopen(str,"rb"))==NULL) {
	 printf("Cannot open %s\n",str);
	 exit(-1);
      }
      n2=2*wfpc2_n2psf[i]+1;
      n3=2*WFPC2_RPSF+1;
      n3skip=wfpc2_rpsf[i]-WFPC2_RPSF;
      wfpc2psflib[i][j]=(float******)calloc(sizeof(float*****),wfpc2_nypsfpos[i]);
      ptr1=(float*****)calloc(sizeof(float****),wfpc2_nypsfpos[i]*wfpc2_nxpsfpos[i]);
      ptr2=(float****)calloc(sizeof(float***),wfpc2_nypsfpos[i]*wfpc2_nxpsfpos[i]*n2);
      ptr3=(float***)calloc(sizeof(float**),wfpc2_nypsfpos[i]*wfpc2_nxpsfpos[i]*n2*n2);
      ptr4=(float**)calloc(sizeof(float*),wfpc2_nypsfpos[i]*wfpc2_nxpsfpos[i]*n2*n2*n3);
      ptr5=(float*)calloc(sizeof(float),wfpc2_nypsfpos[i]*wfpc2_nxpsfpos[i]*n2*n2*n3*n3);
      if (!wfpc2psflib[i][j] || !ptr1) merr();
      for (y1=0;y1<wfpc2_nypsfpos[i];y1++) {
	 wfpc2psflib[i][j][y1]=ptr1;
	 ptr1+=wfpc2_nxpsfpos[i];
	 for (x1=0;x1<wfpc2_nxpsfpos[i];x1++) {
	    wfpc2psflib[i][j][y1][x1]=ptr2+wfpc2_n2psf[i];
	    ptr2+=n2;
	    for (y2=-wfpc2_n2psf[i];y2<=wfpc2_n2psf[i];y2++) {
	       wfpc2psflib[i][j][y1][x1][y2]=ptr3+wfpc2_n2psf[i];
	       ptr3+=n2;
	       for (x2=-wfpc2_n2psf[i];x2<=wfpc2_n2psf[i];x2++) {
		  wfpc2psflib[i][j][y1][x1][y2][x2]=ptr4+WFPC2_RPSF;
		  ptr4+=n3;
		  if (n3skip) fseek(f,4*n3skip*(2*wfpc2_rpsf[i]+1),SEEK_CUR);
		  for (y3=-WFPC2_RPSF;y3<=WFPC2_RPSF;y3++) {
		     wfpc2psflib[i][j][y1][x1][y2][x2][y3]=ptr5+WFPC2_RPSF;
		     ptr5+=n3;
		     if (n3skip) fseek(f,4*n3skip,SEEK_CUR);
		     ffread(wfpc2psflib[i][j][y1][x1][y2][x2][y3]-WFPC2_RPSF,4,n3,f);
		     if (n3skip) fseek(f,4*n3skip,SEEK_CUR);
		  }
		  if (n3skip) fseek(f,4*n3skip*(2*wfpc2_rpsf[i]+1),SEEK_CUR);
	       }
	    }
	 }
      }
      fclose(f);
   }
   return;
}

void wfpc2freepsf(void) {
   int i,j;
   for (i=0;i<3;i++) for (j=0;j<WFPC2_NFILTERS;j++) if (wfpc2psflib[i][j]) {
      free(wfpc2psflib[i][j][0][0][-wfpc2_n2psf[i]][-wfpc2_n2psf[i]][-WFPC2_RPSF]-WFPC2_RPSF);
      free(wfpc2psflib[i][j][0][0][-wfpc2_n2psf[i]][-wfpc2_n2psf[i]]-WFPC2_RPSF);
      free(wfpc2psflib[i][j][0][0][-wfpc2_n2psf[i]]-wfpc2_n2psf[i]);
      free(wfpc2psflib[i][j][0][0]-wfpc2_n2psf[i]);
      free(wfpc2psflib[i][j][0]);
      free(wfpc2psflib[i][j]);
      wfpc2psflib[i][j]=NULL;
   }
   return;
}

extern int lastpsftype;
extern float **psf;

/*
void calcwfpc2psf(int img,float x,float y,int r) {
   int i,j,y1,x1,y2,x2,yy,xx;
   float mx1,my1,imx1,imy1;
   float mx2,my2,imx2,imy2;
   static int first=1,lastr=0,lastimg=0;
   static float lastx=0,lasty=0;

   if (!first && lastpsftype==1 && img==lastimg && x==lastx && y==lasty && r<=lastr) return;
   first=0;
   lastpsftype=1;
   lastimg=img;
   lastx=x;
   lasty=y;
   lastr=r;
   i=wfpc2psflu[img][0];
   j=wfpc2psflu[img][1];
   //printf("%s/%s PSF at %f,%f; r=%d:\n",wfpc2_cn[i],WFPC2filters[j].name,x,y,r);
   y1=(int)(y-128)/256; if (y1<0) y1=0; if (y1>=wfpc2_nypsfpos[i]-1) y1=wfpc2_nypsfpos[i]-2;
   imy1=(y-128-256*y1)/256.; my1=1-imy1;
   x1=(int)(x-128)/256; if (x1<0) x1=0; if (x1>=wfpc2_nxpsfpos[i]-1) x1=wfpc2_nxpsfpos[i]-2;
   imx1=(x-128-256*x1)/256.; mx1=1-imx1;
   y=(y-(int)(y+50)+49.5)*wfpc2_sub[i];
   y2=(int)(y+wfpc2_sub[i])-wfpc2_sub[i];
   imy2=y-y2; my2=1-imy2;
   x=(x-(int)(x+50)+49.5)*wfpc2_sub[i];
   x2=(int)(x+wfpc2_sub[i])-wfpc2_sub[i];
   imx2=x-x2; mx2=1-imx2;
   //printf("y1=%d, my1=%f; x1=%d, mx1=%f\n",y1,my1,x1,mx1);
   //printf("y2=%d, my2=%f; x2=%d, mx2=%f\n",y2,my2,x2,mx2);
   for (yy=-r;yy<=r;yy++) for (xx=-r;xx<=r;xx++) {
      psf[yy][xx]=0.;
      psf[yy][xx]+=(wfpc2psflib[i][j][y1][x1][y2][x2][yy][xx]*mx2*my2+wfpc2psflib[i][j][y1][x1][y2][x2+1][yy][xx]*imx2*my2+wfpc2psflib[i][j][y1][x1][y2+1][x2][yy][xx]*mx2*imy2+wfpc2psflib[i][j][y1][x1][y2+1][x2+1][yy][xx]*imx2*imy2)*mx1*my1;
      psf[yy][xx]+=(wfpc2psflib[i][j][y1][x1+1][y2][x2][yy][xx]*mx2*my2+wfpc2psflib[i][j][y1][x1+1][y2][x2+1][yy][xx]*imx2*my2+wfpc2psflib[i][j][y1][x1+1][y2+1][x2][yy][xx]*mx2*imy2+wfpc2psflib[i][j][y1][x1+1][y2+1][x2+1][yy][xx]*imx2*imy2)*imx1*my1;
      psf[yy][xx]+=(wfpc2psflib[i][j][y1+1][x1][y2][x2][yy][xx]*mx2*my2+wfpc2psflib[i][j][y1+1][x1][y2][x2+1][yy][xx]*imx2*my2+wfpc2psflib[i][j][y1+1][x1][y2+1][x2][yy][xx]*mx2*imy2+wfpc2psflib[i][j][y1+1][x1][y2+1][x2+1][yy][xx]*imx2*imy2)*mx1*imy1;
      psf[yy][xx]+=(wfpc2psflib[i][j][y1+1][x1+1][y2][x2][yy][xx]*mx2*my2+wfpc2psflib[i][j][y1+1][x1+1][y2][x2+1][yy][xx]*imx2*my2+wfpc2psflib[i][j][y1+1][x1+1][y2+1][x2][yy][xx]*mx2*imy2+wfpc2psflib[i][j][y1+1][x1+1][y2+1][x2+1][yy][xx]*imx2*imy2)*imx1*imy1;
   }
   //for (yy=6;yy>=-6;yy--) if (yy>=-r && yy<=r) {for (xx=-6;xx<=6;xx++) if (xx>=-r && xx<=r) printf("%5d ",(int)(psf[yy][xx]*100000+0.5)); printf("\n");} fflush(stdout);
   return;
}
*/

int calcwfpc2psf(int img,float x,float y,int r) {
   int i,j,y1,x1,y2,x2,yy,xx;
   float mx2,my2,imx2,imy2;
   static int first=1,lastr=0,lastimg=0;
   static float lastx=0,lasty=0;
   extern int poffreset;

   if (!first && lastpsftype==1 && img==lastimg && x==lastx && y==lasty && r<=lastr && !poffreset) return 0;
   first=0;lastpsftype=1;lastimg=img;lastx=x;lasty=y;lastr=r;
   i=wfpc2psflu[img][0];
   j=wfpc2psflu[img][1];
   //printf("%d %d %d %f %f %d",img,i,j,x,y,r); fflush(stdout);
   //printf("%s/%s PSF at %f,%f; r=%d:\n",wfpc2_cn[i],WFPC2filters[j].name,x,y,r);
   if (img==Nimg && WFPC2_DRIZZLE_BASE) {
      y1=4;
      x1=7;
   }
   else {
      y1=(int)y/100; if (y1<0) y1=0; if (y1>=wfpc2_nypsfpos[i]) y1=wfpc2_nypsfpos[i]-1;
      x1=(int)x/100; if (x1<0) x1=0; if (x1>=wfpc2_nxpsfpos[i]) x1=wfpc2_nxpsfpos[i]-1;
   }
   y-=(int)y; if (y<0) y++;
   y=(y-0.5)*wfpc2_sub[i];
   y2=(int)(y+wfpc2_sub[i])-wfpc2_sub[i];
   imy2=y-y2; my2=1-imy2;
   //printf(" (%f %d %d %f %f)",y,y1,y2,my2,imy2); fflush(stdout);
   x-=(int)x; if (x<0) x++;
   x=(x-0.5)*wfpc2_sub[i];
   x2=(int)(x+wfpc2_sub[i])-wfpc2_sub[i];
   imx2=x-x2; mx2=1-imx2;
   //printf(" (%f %d %d %f %f)\n",x,x1,x2,mx2,imx2); fflush(stdout);
   //printf("y1=%d, my1=%f; x1=%d, mx1=%f\n",y1,my1,x1,mx1);
   //printf("y2=%d, my2=%f; x2=%d, mx2=%f\n",y2,my2,x2,mx2);
   for (yy=-r;yy<=r;yy++) for (xx=-r;xx<=r;xx++) psf[yy][xx]=wfpc2psflib[i][j][y1][x1][y2][x2][yy][xx]*mx2*my2+wfpc2psflib[i][j][y1][x1][y2][x2+1][yy][xx]*imx2*my2+wfpc2psflib[i][j][y1][x1][y2+1][x2][yy][xx]*mx2*imy2+wfpc2psflib[i][j][y1][x1][y2+1][x2+1][yy][xx]*imx2*imy2;
   //for (yy=6;yy>=-6;yy--) if (yy>=-r && yy<=r) {for (xx=-6;xx<=6;xx++) if (xx>=-r && xx<=r) printf("%5d ",(int)(psf[yy][xx]*100000+0.5)); printf("\n");} fflush(stdout);
   return 1;
}

double WFPC2calcmag(int img,float x0,float y0,float ct0,float bg,int useCTE) {
   float x,y,cm=1.;
   double m;

   if (WFPC2_NFILTERS<0) WFPC2initfilters();
   m=-2.5*log10(ct0*apcor[img]/iEXP[img]*wfpc2_ctmult[wfpc2psflu[img][3]][wfpc2psflu[img][0]])+WFPC2filters[wfpc2psflu[img][1]].zp[wfpc2psflu[img][0]];
   if (useCTE) {
      shift(img,x0,y0,&x,&y,1);
      if (iEXP0[img]>0.) cm=iEXP[img]/iEXP0[img];
      m-=WFPC2_CTE(wfpc2psflu[img][0],x,y,ct0,cm,wfpc2psflu[img][3],bg,iEPOCH[img],wfpc2psflu[img][2],wfpc2psflu[img][1]);
   }
   return m;
}

void WFPC2outstarinfo(FILE *f,int*ct) {
   int i,j,n;
   for (i=0;i<WFPC2_NFILTERS;i++) {
      n=0;
      for (j=0;j<Nimg;j++) if (wfpc2psflu[j][1]==i) n++;
      if (n>1) {
	 fprintf(f,"%d. Total counts, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Total sky level, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Instrumental VEGAMAG magnitude, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Transformed UBVRI magnitude, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Magnitude uncertainty, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Chi, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Signal-to-noise, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Sharpness, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Roundness, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Crowding, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. PSF FWHM, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. PSF eccentricity, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. PSF a parameter, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. PSF b parameter, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. PSF c parameter, %s\n",++(*ct),WFPC2filters[i].name);
	 fprintf(f,"%d. Photometry quality flag, %s\n",++(*ct),WFPC2filters[i].name);
      }
   }
}

typedef struct{float ct0,ct,ctcorr,sky,m,dm,chi,sn,sh,rnd,r,e,a,b,c,crowd;int flag;} photdatatype;
void WFPC2outstar(FILE *of,float x0,float y0,photdatatype*pdata) {
   int i,j;
   float cm=1.;
   static int *fused=NULL;
   static photdatatype*fphot=NULL;
   static double*smag=NULL,*vmag=NULL,*dvmag=NULL;
   float x,y;

   if (WFPC2_NFILTERS<0) WFPC2initfilters();
   if (!fphot) {
      fused=(int*)calloc(sizeof(int),WFPC2_NFILTERS);
      fphot=(photdatatype*)calloc(sizeof(photdatatype),WFPC2_NFILTERS);
      smag=(double*)calloc(sizeof(double),WFPC2_NFILTERS);
      vmag=(double*)calloc(sizeof(double),WFPC2_NFILTERS);
      dvmag=(double*)calloc(sizeof(double),WFPC2_NFILTERS);
      if (!fused || !fphot || !smag || !vmag || !dvmag) merr();
   }
   for (j=0;j<Nimg;j++) {
      double dm = -2.5*log10(wfpc2_ctmult[wfpc2psflu[j][3]][wfpc2psflu[j][0]]) + WFPC2filters[wfpc2psflu[j][1]].zp[wfpc2psflu[j][0]]-Zero;
      if (pdata[j].ct>0) {
	 if (UseCTE) {
	    shift(j,x0,y0,&x,&y,1);
	    if (iEXP0[j]>0.) cm=iEXP[j]/iEXP0[j];
	    dm -= WFPC2_CTE(wfpc2psflu[j][0],x,y,pdata[j].ct,cm,wfpc2psflu[j][3],pdata[j].sky,iEPOCH[j],wfpc2psflu[j][2],wfpc2psflu[j][1]);
	 }
	 pdata[j].m += dm;
      }
      pdata[j].ctcorr *= pow(10,-0.4*dm);
   }
   for (i=0;i<WFPC2_NFILTERS;i++) {
      float wt,swt,twt=0,tswt=0,S=0.,SS=0.,is,iss,tcm=0.;
      fphot[i].ct0=fphot[i].ct=fphot[i].sky=fphot[i].dm=fphot[i].chi=fphot[i].sn=fphot[i].sh=fphot[i].rnd=fphot[i].r=fphot[i].e=fphot[i].a=fphot[i].b=fphot[i].c=fphot[i].m=fphot[i].crowd=0.;
      fused[i]=0;
      fphot[i].flag=0;
      for (j=0;j<Nimg;j++) if (wfpc2psflu[j][1]==i) {
	 fused[i]++;
	 if (pdata[j].flag<8 && !(pdata[j].flag&FlagMask)) {
	    is=pdata[j].ct/iEXP[j];
	    iss=is/pdata[j].sn;
	    wt=1./iss/iss;
	    if (is>0) swt=wt*is;
	    else swt=0.0;
	    twt+=wt;
	    tswt+=swt;
	    fphot[i].ct0+=pdata[j].ct0/iEXP[j]*wt;
	    fphot[i].ct+=pdata[j].ct/iEXP[j]*wt;
	    fphot[i].chi+=pdata[j].chi*pdata[j].chi*wt;
	    fphot[i].sh+=pdata[j].sh*swt;
	    fphot[i].sky+=pdata[j].sky/iEXP[j]*wt;
	    fphot[i].m+=pdata[j].ctcorr*wt;
	    fphot[i].rnd+=pdata[j].rnd*swt;
	    fphot[i].crowd+=pdata[j].crowd*wt;
	    fphot[i].r+=pdata[j].r*swt;
	    fphot[i].e+=pdata[j].e*swt;
	    fphot[i].a+=pdata[j].a*swt;
	    fphot[i].b+=pdata[j].b*swt;
	    fphot[i].c+=pdata[j].c*swt;
	    fphot[i].flag|=pdata[j].flag;
	    S+=is*wt;
	    SS+=iss*iss*wt*wt;
	 }
	 tcm+=iEXP[j];
      }
      if (twt>0.) {
	 fphot[i].ct0/=twt/tcm;
	 fphot[i].ct/=twt/tcm;
	 fphot[i].chi=sqrt(fphot[i].chi/twt);
	 fphot[i].sky/=twt/tcm;
	 if (fphot[i].m>0) fphot[i].m=-2.5*log10(fphot[i].m/twt);
	 else fphot[i].m=99.999;
	 fphot[i].crowd/=twt;
	 S/=twt;
	 SS=sqrt(SS)/twt;
	 fphot[i].sn=S/SS;
	 fphot[i].dm=1.0857362*SS/S;
      }
      else fphot[i].m=99.999;
      if (tswt>0.) {
	 fphot[i].sh/=tswt;
	 fphot[i].rnd/=tswt;
	 fphot[i].r/=tswt;
	 fphot[i].e/=tswt;
	 fphot[i].a/=tswt;
	 fphot[i].b/=tswt;
	 fphot[i].c/=tswt;
      }
      vmag[i]=fphot[i].m;
      dvmag[i]=fphot[i].dm;
   }
   WFPC2transform(vmag,dvmag,smag);
   for (i=0;i<WFPC2_NFILTERS;i++) if (fused[i]>1) {
      if (fphot[i].ct<999999.5) fprintf(of,"  %8.1f",fphot[i].ct);
      else fprintf(of,"  %8.2e",fphot[i].ct);
      if (fphot[i].sky<99999.95) fprintf(of," %8.2f",fphot[i].sky);
      else if (fphot[i].sky<999999.5) fprintf(of," %8.1f",fphot[i].sky);
      else fprintf(of," %8.1e",fphot[i].sky);
      fprintf(of," %6.3f %6.3f",fphot[i].m,smag[i]);
      if (fphot[i].dm>9.999) fprintf(of," 9.999");
      else fprintf(of," %5.3f",fphot[i].dm);
      fprintf(of," %6.2f %7.1f %6.3f %6.3f %5.3f %6.3f %6.3f %6.3f %6.3f %6.3f %2d",fphot[i].chi,fphot[i].sn,fphot[i].sh,fphot[i].rnd,fphot[i].crowd,fphot[i].r,fphot[i].e,fphot[i].a,fphot[i].b,fphot[i].c,fphot[i].flag);
   }
   for (j=0;j<Nimg;j++) {
      if (pdata[j].ct<999999.5) fprintf(of,"  %8.1f",pdata[j].ct);
      else fprintf(of,"  %8.2e",pdata[j].ct);
      if (pdata[j].sky<99999.95) fprintf(of," %8.2f",pdata[j].sky);
      else if (pdata[j].sky<999999.5) fprintf(of," %8.1f",pdata[j].sky);
      else fprintf(of," %8.1e",pdata[j].sky);
      if (pdata[j].ct<=0) x=pdata[j].m;
      else if (smag[wfpc2psflu[j][1]]>99) x=smag[wfpc2psflu[j][1]];
      else x=pdata[j].m+smag[wfpc2psflu[j][1]]-vmag[wfpc2psflu[j][1]];
      fprintf(of," %6.3f %6.3f",pdata[j].m,x);
      if (pdata[j].dm>9.999) fprintf(of," 9.999");
      else fprintf(of," %5.3f",pdata[j].dm);
      fprintf(of," %6.2f %7.1f %6.3f %6.3f %5.3f %6.3f %6.3f %6.3f %6.3f %6.3f %2d",pdata[j].chi,pdata[j].sn,pdata[j].sh,pdata[j].rnd,pdata[j].crowd,pdata[j].r,pdata[j].e,pdata[j].a,pdata[j].b,pdata[j].c,pdata[j].flag);
   }
   return;
}

float wfpc2_apsize(int img,float x,float y) {
   double area;
   int cm=wfpc2psflu[img][0];
   if (x<0) x=0;
   else if (x>800) x=800;
   if (y<0) y=0;
   else if (y>800) y=800;
   area = WFPC2pixsize(cm,x,y);
   //printf("%d %f %f %f\n",cm,x,y,area);
   return 10.979/sqrt(area);
}

void WFPC2shift(int img,float*x,float*y) {
   double xx,yy;
   xx=*x; yy=*y;
   if (wfpc2psflu[img][0]>=0) WFPC2fwddistort(wfpc2psflu[img][0],&xx,&yy);
   *x=xx; *y=yy;
   return;
}

void WFPC2unshift(int img,float*x,float*y) {
   double xx,yy;
   xx=*x; yy=*y;
   if (wfpc2psflu[img][0]>=0) WFPC2revdistort(wfpc2psflu[img][0],&xx,&yy);
   *x=xx; *y=yy;
   return;
}

extern FILE *finfo;
void writewfpc2info(void) {
   int img;
   fprintf(finfo,"* WFPC2-specific info\n");
   for (img=0;img<Nimg;img++) {
      fprintf(finfo,"* %s %d %f\n",WFPC2filters[wfpc2psflu[img][1]].name,wfpc2psflu[img][0],iEXP[img]);
   }
   return;
}

static int*ffused=NULL;
static double*fakem0;
void WFPC2readfakemag(FILE*f) {
   int i,img;

   if (WFPC2_NFILTERS<0) WFPC2initfilters();
   if (ffused==NULL) {
      ffused=(int*)calloc(sizeof(int),WFPC2_NFILTERS);
      fakem0=(double*)calloc(sizeof(double),WFPC2_NFILTERS);
      if (!ffused || !fakem0) merr();
      for (img=0;img<Nimg;img++) ffused[wfpc2psflu[img][1]]++;
   }
   for (i=0;i<WFPC2_NFILTERS;i++) if (ffused[i]) fscanf(f,"%lf",fakem0+i);
   return;
}

void WFPC2fixfakemag(int img,float x0,float y0,double*ct0,float*bg) {
   int i;
   double dm;

   dm=WFPC2calcmag(img,x0,y0,1.0,bg[img],0)-fakem0[wfpc2psflu[img][1]];
   ct0[img]=pow(10,0.4*dm);
   for (i=0;i<3;i++) {
      dm=WFPC2calcmag(img,x0,y0,ct0[img],bg[img],UseCTE)-fakem0[wfpc2psflu[img][1]];
      ct0[img]*=pow(10,0.4*dm);
   }
   return;
}
